package com.zhili.stationserver.component.stationserver.handler;

import com.alibaba.fastjson.JSON;
import com.zhili.stationserver.annotation.ProtocalTranform;
import com.zhili.stationserver.annotation.TransformField;
import com.zhili.stationserver.component.ApplicationContextComponent;
import com.zhili.stationserver.component.stationserver.StationCenter;
import com.zhili.stationserver.configuration.ServerConfig;
import com.zhili.stationserver.constant.TransformType;
import com.zhili.stationserver.dto.*;
import com.zhili.stationserver.util.BytesUtil;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.ChannelInboundHandlerAdapter;
import io.netty.handler.timeout.IdleState;
import io.netty.handler.timeout.IdleStateEvent;
import io.netty.util.AttributeKey;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.data.redis.core.script.DefaultRedisScript;
import org.springframework.data.redis.core.script.RedisScript;
import org.springframework.kafka.core.KafkaTemplate;

import java.io.UnsupportedEncodingException;
import java.lang.reflect.Field;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;

/**
 * @author ：HuangBin
 * @description：TODO
 * @date ：2022/12/20 15:40
 */
@Slf4j
public class StationMessageHandler extends ChannelInboundHandlerAdapter {
    KafkaTemplate kafkaTemplate = (KafkaTemplate) ApplicationContextComponent.getApplicationContext().getBean("kafkaTemplate");
    ServerConfig serverConfig = ApplicationContextComponent.getApplicationContext().getBean(ServerConfig.class);
    RedisTemplate redisTemplate = (RedisTemplate) ApplicationContextComponent.getApplicationContext().getBean("redisTemplate");
    RedisScript stationRegisterScript = (DefaultRedisScript) ApplicationContextComponent.getApplicationContext().getBean("stationRegisterScript");

    //检测心跳超时
    @Override
    public void userEventTriggered(ChannelHandlerContext ctx, Object evt) throws Exception {
        super.userEventTriggered(ctx, evt);
        if (evt instanceof IdleStateEvent) {
            IdleState state = ((IdleStateEvent) evt).state();
            if (state == IdleState.READER_IDLE) {
                ctx.channel().close();
                log.error("-------客户端心跳超时");
            }
        }
    }


    @Override
    public void channelActive(ChannelHandlerContext ctx) throws Exception {
        log.info("channel connected: " + ctx.channel().remoteAddress());
        StationCenter.add(ctx.channel());
        super.channelActive(ctx);
    }

    @Override
    public void channelInactive(ChannelHandlerContext ctx) throws Exception {
        log.info("channel disconnected: " + ctx.channel().remoteAddress());
        super.channelInactive(ctx);
        //断连后及时删除站的注册信息
        Object stationCodeObj = ctx.channel().attr(AttributeKey.valueOf("station")).get();
        if (stationCodeObj == null || StringUtils.isBlank(stationCodeObj.toString())) {
            return;
        }
        redisTemplate.delete("StationConn:" + stationCodeObj);
    }

    @Override
    public void channelRead(ChannelHandlerContext ctx, Object msg) throws Exception {
        if (!(msg instanceof StationMessage)) {
            return;
        }
        StationMessage stationMessage = (StationMessage) msg;
        Integer command = stationMessage.getCommand();
        byte[] data = stationMessage.getData();
        //log.info("command:" + command + ",data:" + ArrayUtils.toString(data));
        if (command.equals(StationMessage.CommandType.SIGN_IN)) {
            //签入
            byte[] stationCodeBytes = ArrayUtils.subarray(data, 0, 8);
            String stationCode = BytesUtil.parseString(stationCodeBytes);
            log.info("SIGN_IN:" + stationCode);
            StationConnInfo stationConnInfo = new StationConnInfo();
            stationConnInfo.setServerInstanceId(serverConfig.getInstanceId());
            stationConnInfo.setConnIdStr(StationCenter.getChannelIdStr(ctx.channel()));
            Long success = register(stationCode, stationConnInfo);
            log.info("签入:{}", success);
            StationMessage stationMessageReply = new StationMessage();
            stationMessageReply.setCommand(StationMessage.CommandType.SIGN_IN_REPLY);
            byte[] replyBytes = new byte[9];
            BytesUtil.setSubBytes(replyBytes, 0, stationCodeBytes);
            if (success.equals(0l)) {
                //成功
                log.info("签入成功");
                StationCenter.markStation(ctx.channel(), stationCode);
                replyBytes[8] = (byte) 0;
                stationMessageReply.setData(replyBytes);
                ctx.channel().writeAndFlush(stationMessageReply);
            } else {
                log.info("签入失败");
                replyBytes[8] = (byte) 1;
                stationMessageReply.setData(replyBytes);
                ctx.channel().writeAndFlush(stationMessageReply);
                ctx.channel().close();
            }

        } else if (command.equals(StationMessage.CommandType.HEART_BEAT)) {
            //心跳
            String stationCode = StationCenter.getStation(ctx.channel());
            if (StringUtils.isBlank(stationCode)) {
                //没有签入
                ctx.channel().close();
                return;
            }
            StationConnInfo stationConnInfo = new StationConnInfo();
            stationConnInfo.setServerInstanceId(serverConfig.getInstanceId());
            stationConnInfo.setConnIdStr(StationCenter.getChannelIdStr(ctx.channel()));
            Long success = register(stationCode, stationConnInfo);
            //log.info("心跳:{}", success);
            StationMessage stationMessageReply = new StationMessage();
            stationMessageReply.setCommand(StationMessage.CommandType.HEART_BEAT_REPLY);
            byte[] replyBytes = new byte[1];
            if (success.equals(0l)) {
                //成功
                StationCenter.markStation(ctx.channel(), stationCode);
                replyBytes[0] = (byte) 0;
                stationMessageReply.setData(replyBytes);
                ctx.channel().writeAndFlush(stationMessageReply);
            } else {
                replyBytes[0] = (byte) 1;
                stationMessageReply.setData(replyBytes);
                ctx.channel().writeAndFlush(stationMessageReply);
                ctx.channel().close();
            }

        } else if (command.equals(StationMessage.CommandType.PERCEPTION_VEH)) {
            //车辆感知
            String stationCode = StationCenter.getStation(ctx.channel());
            if (StringUtils.isBlank(stationCode)) {
                return;
            }
            PerceptionVehInfoDto dto = parse(stationMessage, PerceptionVehInfoDto.class);
            dto.setStationCode(stationCode);
            String topic = serverConfig.getServiceCenterName() + "_PERCEPTION_VEH";
            kafkaTemplate.send(topic, JSON.toJSONString(dto));
        } else if (command.equals(StationMessage.CommandType.PERCEPTION_BAT_AVAILABLE)) {
            //可用电池数感知
            String stationCode = StationCenter.getStation(ctx.channel());
            if (StringUtils.isBlank(stationCode)) {
                return;
            }
            //log.info("================可用电池数HASH：{}",BytesUtil.bytesToHexString(stationMessage.getData()));
            PerceptionUploadAvailableBatsDto dto = parse(stationMessage, PerceptionUploadAvailableBatsDto.class);
            dto.setStationCode(stationCode);
            String topic = serverConfig.getServiceCenterName() + "_PERCEPTION_BAT_AVAILABLE";
            //log.info("================可用电池数：{}",dto);
            kafkaTemplate.send(topic, JSON.toJSONString(dto));
        } else if (command.equals(StationMessage.CommandType.PERCEPTION_PLC)) {
            //plc感知
            String stationCode = StationCenter.getStation(ctx.channel());
            if (StringUtils.isBlank(stationCode)) {
                return;
            }
            PlcPerceptionInfoDto dto = parse(stationMessage, PlcPerceptionInfoDto.class);
            dto.setStationCode(stationCode);
            String topic = serverConfig.getServiceCenterName() + "_PERCEPTION_PLC";
            kafkaTemplate.send(topic, JSON.toJSONString(dto));
        } else if (command.equals(StationMessage.CommandType.PERCEPTION_EXCHANGE_INFO)) {
            //换电信息感知
            String stationCode = StationCenter.getStation(ctx.channel());
            if (StringUtils.isBlank(stationCode)) {
                return;
            }
            PerceptionExchangeInfoDto dto = parse(stationMessage, PerceptionExchangeInfoDto.class);
            dto.setStationCode(stationCode);
            String topic = serverConfig.getServiceCenterName() + "_PERCEPTION_EXCHANGE_INFO";
            kafkaTemplate.send(topic, JSON.toJSONString(dto));
        } else if (command.equals(StationMessage.CommandType.PERCEPTION_EXCHANGE_LOG)) {
            String stationCode = StationCenter.getStation(ctx.channel());
            if (StringUtils.isBlank(stationCode)) {
                return;
            }
            ExchangeLogDto dto = parse(stationMessage, ExchangeLogDto.class);
            dto.setStationCode(stationCode);
            String topic = serverConfig.getServiceCenterName() + "_EXCHANGE_LOG";
            kafkaTemplate.send(topic, JSON.toJSONString(dto));
        } else if (command.equals(StationMessage.CommandType.REPLY_EXCHANGE)) {
            String stationCode = StationCenter.getStation(ctx.channel());
            if (StringUtils.isEmpty(stationCode)) {
                return;
            }
            StartEventReplyDTO startSwapReply = parse(stationMessage, StartEventReplyDTO.class);
            startSwapReply.setStationCode(stationCode);
            BaseControlReplyDTO baseControlReplyDTO = new BaseControlReplyDTO();
            baseControlReplyDTO.setType("start_reply");
            baseControlReplyDTO.setData(startSwapReply);
            String topic = serverConfig.getServiceCenterName() + "_CONTROL_REPLY";
            log.info("启动换电回复：{}", baseControlReplyDTO);
            kafkaTemplate.send(topic, JSON.toJSONString(baseControlReplyDTO));
        } else if (command.equals(StationMessage.CommandType.EXCHANGE_RECORD)) {
            log.info("上传换电记录：{}", data);
            SwapRecordDTO exchangeRecordInfo = parse(stationMessage, SwapRecordDTO.class);
            if (exchangeRecordInfo.getStartSource() == 0) {
                exchangeRecordInfo.setDriverId(null);
            }
            String topic = serverConfig.getServiceCenterName() + "_UPLOAD_EXCHANGE_RECORD";
            exchangeRecordInfo.setServiceInstanceId(serverConfig.getInstanceId());
            exchangeRecordInfo.setConnIdStr(StationCenter.getChannelIdStr(ctx.channel()));
            log.info("换电记录上传：{}", exchangeRecordInfo);
            kafkaTemplate.send(topic, JSON.toJSONString(exchangeRecordInfo));
        } else if (command.equals(StationMessage.CommandType.PERCEPTION_CHARGE_RECORD)) {
            String stationCode = StationCenter.getStation(ctx.channel());
            if (StringUtils.isBlank(stationCode)) {
                return;
            }
            PerceptionChargeRecordDto dto = parse(stationMessage, PerceptionChargeRecordDto.class);
            dto.setStationCode(stationCode);
            dto.setServiceInstanceId(serverConfig.getInstanceId());
            dto.setConnIdStr(StationCenter.getChannelIdStr(ctx.channel()));
            log.info("上传充电记录：{}", dto);
            String topic = serverConfig.getServiceCenterName() + "_UPLOAD_CHARGE_RECORD";
            kafkaTemplate.send(topic, JSON.toJSONString(dto));
        } else if (command.equals(StationMessage.CommandType.SET_CHARGE_RATE_REPLY)) {
            String stationCode = StationCenter.getStation(ctx.channel());
            if (StringUtils.isBlank(stationCode)) {
                return;
            }
            ChargeRateReplyDto replyDto = parse(stationMessage, ChargeRateReplyDto.class);
            log.info("下发费率回复：{}", replyDto);
            replyDto.setStationCode(stationCode);
            String topic = serverConfig.getServiceCenterName() + "_SET_RATE_REPLY";
            kafkaTemplate.send(topic, JSON.toJSONString(replyDto));
        } else if (command.equals(StationMessage.CommandType.PERCEPTION_BAT_GUN_STATE)) {
            String stationCode = StationCenter.getStation(ctx.channel());
            if (StringUtils.isBlank(stationCode)) {
                return;
            }
            BatGunStateDto dto = parse(stationMessage, BatGunStateDto.class);
            //log.info("站控上传插枪信息：{}", dto);
            dto.setStationCode(stationCode);
            String topic = serverConfig.getServiceCenterName() + "_PERCEPTION_BAT_GUN_STATE";
            kafkaTemplate.send(topic, JSON.toJSONString(dto));
        }else if (command.equals(StationMessage.CommandType.START_CHARGE_REPLY)) {
            String stationCode = StationCenter.getStation(ctx.channel());
            if (StringUtils.isBlank(stationCode)) {
                return;
            }
            ChargeStartReply chargeStartReply = parse(stationMessage, ChargeStartReply.class);
            log.info("小程序扫码充电回复：{}",chargeStartReply);
            chargeStartReply.setStationCode(stationCode);
            chargeStartReply.setType("startChargeReply");
            String topic = serverConfig.getServiceCenterName() + "_CHARGER_REPLY";
            kafkaTemplate.send(topic, JSON.toJSONString(chargeStartReply));
        }else if (command.equals(StationMessage.CommandType.STOP_CHARGE_REPLY)) {
            String stationCode = StationCenter.getStation(ctx.channel());
            if (StringUtils.isBlank(stationCode)) {
                return;
            }
            ChargeStopReply chargeStopReply = parse(stationMessage, ChargeStopReply.class);
            log.info("小程序断电回复：{}",chargeStopReply);
            chargeStopReply.setStationCode(stationCode);
            chargeStopReply.setType("stopChargeReply");
            String topic = serverConfig.getServiceCenterName() + "_CHARGER_REPLY";
            kafkaTemplate.send(topic, JSON.toJSONString(chargeStopReply));
        }else if (command.equals(StationMessage.CommandType.AUTO_CHARGE_REPLY)) {
            String stationCode = StationCenter.getStation(ctx.channel());
            if (StringUtils.isBlank(stationCode)) {
                return;
            }
            AutoChargeReply autoChargeReply = parse(stationMessage, AutoChargeReply.class);
            log.info("小程自动切枪电回复：{}",autoChargeReply);
            autoChargeReply.setStationCode(stationCode);
            autoChargeReply.setType("autoChargeReply");
            String topic = serverConfig.getServiceCenterName() + "_CHARGER_REPLY";
            kafkaTemplate.send(topic, JSON.toJSONString(autoChargeReply));
        }else if (command.equals(StationMessage.CommandType.STATION_INFO)) {
            //站点信息
            StationInfoDto dto = parse(stationMessage, StationInfoDto.class);
            dto.setStationCode(dto.getStationCode());
            String topic = serverConfig.getServiceCenterName() + "_STATION_INFO";
            kafkaTemplate.send(topic, JSON.toJSONString(dto));
        }
    }

    Long register(String stationCode, StationConnInfo connInfo) {
        List<String> stationRegisterkeys = new ArrayList();
        stationRegisterkeys.add("StationConn:" + stationCode);
        //log.info("sign in stationCode:{}", stationCode);
        return (Long) redisTemplate.execute(stationRegisterScript, stationRegisterkeys, connInfo, 15);
    }

    @Override
    public void exceptionCaught(ChannelHandlerContext ctx, Throwable cause) throws Exception {
        super.exceptionCaught(ctx, cause);
        cause.printStackTrace();
        log.error(cause.getMessage());
    }

    /**
     * 保留指定位数float值
     *
     * @param number
     * @param decimalPlaces
     * @return
     */
    public float formatFloat(float number, int decimalPlaces) {
        DecimalFormatSymbols symbols = new DecimalFormatSymbols(Locale.US);
        DecimalFormat decimalFormat = new DecimalFormat();
        decimalFormat.setDecimalFormatSymbols(symbols);
        decimalFormat.setMaximumFractionDigits(decimalPlaces);
        decimalFormat.setMinimumFractionDigits(decimalPlaces);
        String pattern = "#.";
        for (int i = 0; i < decimalPlaces; i++) {
            pattern += "0";
        }
        decimalFormat.applyPattern(pattern);
        String formattedNumber = decimalFormat.format(number);
        return Float.parseFloat(formattedNumber);
    }

    <T> T parse(StationMessage msg, Class<T> clazz) throws InstantiationException, IllegalAccessException, UnsupportedEncodingException {
        ProtocalTranform annotation = clazz.getAnnotation(ProtocalTranform.class);
        if (annotation == null || annotation.command() != msg.getCommand()) {
            return null;
        }
        byte[] data = msg.getData();
        T t = clazz.newInstance();
        Field[] fields = clazz.getDeclaredFields();
        for (Field field : fields) {
            try {
                TransformField ann = field.getAnnotation(TransformField.class);
                if (ann == null) {
                    continue;
                }
                String type = ann.type();
                int start = ann.start();
                field.setAccessible(true);
                if (type.equals(TransformType.DATE_TYPE)) {
                    byte[] dateBytes = ArrayUtils.subarray(data, start, 8);
                    long l = BytesUtil.bytesToLong(dateBytes);
                    Date date = new Date(l);
                    field.set(t, date);
                } else if (type.equals(TransformType.UINT_TYPE)) {
//                    int length = ann.length();
//                    byte[] uIntBytes = ArrayUtils.subarray(data, start, start + length);
//                    if (uIntBytes != null) {
//                        Integer i = BytesUtil.toIntWithLowerFirst(uIntBytes);
//                        field.set(t, i);
//                    } else {
//                        field.set(t, ann.intDefault());
//                    }
                    int length = ann.length();
                    byte[] uIntBytes = ArrayUtils.subarray(data, start, start + length);
                    int def = ann.intDefault();
                    byte[] defBytes = BytesUtil.fromIntWithLowerFirst(def, length);
                    boolean eq = true;
                    for (int i = 0; i < defBytes.length; i++) {
                        if (defBytes[i] != uIntBytes[i]) {
                            eq = false;
                            break;
                        }
                    }
                    if (eq) {
                        field.set(t, null);
                    } else {
                        Integer i = BytesUtil.toIntWithLowerFirst(uIntBytes);
                        field.set(t, i);
                    }
                } else if (type.equals(TransformType.FLOAT_TYPE)) {
                    int length = ann.length();
                    byte[] uIntBytes = ArrayUtils.subarray(data, start, start + length);
                    byte[] defBytes = BytesUtil.fromIntWithLowerFirst((int) ((ann.floatDefault() - ann.offset()) / ann.accuracy()), length);
                    boolean eq = true;
                    for (int i = 0; i < defBytes.length; i++) {
                        if (defBytes[i] != uIntBytes[i]) {
                            eq = false;
                            break;
                        }
                    }
                    if (eq) {
                        field.set(t, null);
                    } else {
                        Integer i = BytesUtil.toIntWithLowerFirst(uIntBytes);
                        Float res = i * ann.accuracy() + ann.offset();
                        field.set(t, formatFloat(res, ann.floatSignificant()));
                    }
//                    int length = ann.length();
//                    byte[] uIntBytes = ArrayUtils.subarray(data, start, start + length);
//                    if (uIntBytes != null) {
//                        Integer i = BytesUtil.toIntWithLowerFirst(uIntBytes);
//                        Float res = i / ann.accuracy() - ann.offset();
//                        field.set(t, res);
//                    } else {
//                        field.set(t, ann.floatDefault());
//                    }
                } else if (type.equals(TransformType.STRING_TYPE)) {
                    int length = ann.length();
                    byte[] uIntBytes = ArrayUtils.subarray(data, start, start + length);
                    String gbk = new String(uIntBytes, "GBK");
                    if (gbk.indexOf(0) >= 0) {
                        gbk = StringUtils.substring(gbk, 0, gbk.indexOf(0));
                    }
                    field.set(t, gbk);
                } else if (type.equals(TransformType.TIME_STRING_TYPE)) {
                    byte[] tsBytes = ArrayUtils.subarray(data, start, start + 8);
                    String s = BytesUtil.composeTimeString(tsBytes);
//                    log.info("tsBytes:{}", ArrayUtils.toString(tsBytes));
//                    log.info("tString:{}", s);
                    field.set(t, s);
                } else if (type.equals(TransformType.BYTES_TYPE)) {
                    int length = ann.length();
                    byte[] bytes = ArrayUtils.subarray(data, start, start + length);
                    field.set(t, bytes);
                } else if (type.equals(TransformType.CHARGE_LIST_TYPE)) {
                    int length = ann.length();
                    int segmentLen = ann.segmentLen();
                    byte[] bytes = ArrayUtils.subarray(data, start, start + length);
                    int size = length / segmentLen;
                    List<ChargeDetailDto> chargeDetailList = new ArrayList<>();
                    ChargeDetailDto chargeDetailDto;
                    for (int i = 0; i < size; i++) {
                        int index = BytesUtil.toIntWithLowerFirst(ArrayUtils.subarray(bytes, i * segmentLen, i * segmentLen + 1));
                        int power = BytesUtil.toIntWithLowerFirst(ArrayUtils.subarray(bytes, i * segmentLen + 1, i * segmentLen + 3));
                        int startHour = BytesUtil.toIntWithLowerFirst(ArrayUtils.subarray(bytes, i * segmentLen + 3, i * segmentLen + 4));
                        int startMinute = BytesUtil.toIntWithLowerFirst(ArrayUtils.subarray(bytes, i * segmentLen + 4, i * segmentLen + 5));
                        int endHour = BytesUtil.toIntWithLowerFirst(ArrayUtils.subarray(bytes, i * segmentLen + 5, i * segmentLen + 6));
                        int endMinute = BytesUtil.toIntWithLowerFirst(ArrayUtils.subarray(bytes, i * segmentLen + 6, i * segmentLen + 7));
                        int chargeRate = BytesUtil.toIntWithLowerFirst(ArrayUtils.subarray(bytes, i * segmentLen + 7, i * segmentLen + 11));
                        int serRate = BytesUtil.toIntWithLowerFirst(ArrayUtils.subarray(bytes, i * segmentLen + 11, i * segmentLen + 15));
                        int rateType = BytesUtil.toIntWithLowerFirst(ArrayUtils.subarray(bytes, i * segmentLen + 15, i * segmentLen + 1));
                        if (index == 0 && power == 0) {
                            break;
                        }
                        chargeDetailDto = new ChargeDetailDto();
                        chargeDetailDto.setIndex(index);
                        chargeDetailDto.setChargePower(power * 0.01f);
                        chargeDetailDto.setStartHour(startHour);
                        chargeDetailDto.setStartMinute(startMinute);
                        chargeDetailDto.setEndHour(endHour);
                        chargeDetailDto.setEndMinute(endMinute);
                        chargeDetailDto.setChargeRate(chargeRate * 0.000001f);
                        chargeDetailDto.setSerRate(serRate * 0.000001f);
                        chargeDetailDto.setRateType(rateType);
                        chargeDetailList.add(chargeDetailDto);
                    }
                    field.set(t, chargeDetailList);
                }
            } catch (Exception e) {
                e.printStackTrace();
                continue;
            }

        }
        return t;
    }
}
